﻿
' ===========================================================================
'  MODBUS MODULE
' ===========================================================================
Module ModBus

    Friend SentPackets As Int32
    Friend ReceivedPackets As Int32

    ' -------------------------------------------------------------------- GET REGISTER
    Friend Function ModBus_GetRegister(ByVal Device As Byte, _
                                       ByVal StartRegister As UInt16, _
                                       ByVal NumRegisters As UInt16) As Short()
        '
        Dim pv(NumRegisters - 1) As Short
        If SendFc3(Device, StartRegister, NumRegisters, pv) Then
            Return pv
        Else
            ReDim pv(-1)
            Return pv
        End If
    End Function

    ' -------------------------------------------------------------------- WRITE REGISTER
    Friend Sub ModBus_WriteRegister(ByVal Device As Byte, _
                                    ByVal StartRegister As UInt16, _
                                    ByVal NumRegisters As UInt16, _
                                    ByVal Values() As Short)
        '
        SendFc16(Device, StartRegister, NumRegisters, Values)
    End Sub

    ' -------------------------------------------------------------------- IMMEDIATE TO MODBUS
    Friend Sub Modbus_ImmediateToModbus(ByVal Tdata As TransferData)
        With Tdata
            Dim Values(0) As Short
            Values(0) = CShort(.ImmediateValue)
            SendFc16(.Device, .Address, 1, Values)
        End With
    End Sub

    ' -------------------------------------------------------------------- SET DEVICE ID
    Friend Sub Modbus_SetDeviceIdentifier(ByVal ID_Enable As Int32, _
                                          ByVal ID_Address As Int32, _
                                          ByVal ID_Value As Int32)

        FastTimer_Stop()
        '
        Modbus_BroadcastSingleByte(ID_Enable, 0)
        Modbus_BroadcastSingleByte(ID_Address, ID_Value)
        Modbus_BroadcastSingleByte(ID_Enable, 1)
        '
        Threading.Thread.Sleep(250)
        '
        If ModBus_ReadSingleValue(ID_Value, ID_Address) <> ID_Value Then
            MessageBox.Show("ERROR - The device is not working" + vbCrLf + vbCrLf + _
                            " - Test the COM port and the BaudRate" + vbCrLf + _
                            " - Test the power supply and cables" + vbCrLf + _
                            " - Ensure a single device connected" + vbCrLf + _
                            " - Reprogram the device firmware" + vbCrLf + _
                            " - Reinstall the CH340 driver" + vbCrLf + _
                            " - Replace USB cable and connection module", _
                            "Message from Theremino-ModBus", _
                            MessageBoxButtons.OK, MessageBoxIcon.Error)
        Else
            Beep()
        End If
        '
        FastTimer_Start()
    End Sub

    Friend Sub Modbus_BroadcastSingleByte(ByVal address As Int32, _
                                          ByVal value As Int32)
        Dim Values(0) As Short
        Values(0) = CShort(value)
        Dim regindex As UShort = CUShort(DecodeAddress(address.ToString))
        SendFc16(0, regindex, 1, Values)
    End Sub

    Friend Function ModBus_ReadSingleValue(ByVal device As Int32, _
                                           ByVal address As Int32) As Int32
        Dim values(0) As Short
        Dim regindex As UShort = CUShort(DecodeAddress(address.ToString))
        values = ModBus_GetRegister(CByte(device), regindex, 1)
        If values.Length = 1 Then
            Return values(0)
        Else
            Return Integer.MinValue
        End If
    End Function

    ' -------------------------------------------------------------------- Broadcast_Baudrate
    Friend BaudRates(-1) As Int32
    Friend BaudRateRegisterIndex As Int32 = -1
    '
    Friend Sub BroadcastBaudrate(ByVal PortName As String, ByVal baudrate As Int32)
        '
        Dim arrayindex As Int32 = Array.IndexOf(BaudRates, baudrate)
        If arrayindex < 0 Or arrayindex > 255 Then Return
        If BaudRateRegisterIndex < 0 Then Return
        Dim regindex As UShort = CUShort(DecodeAddress(BaudRateRegisterIndex.ToString))
        ' ----------------------------------------------------------------  set baudrate combo items
        SetBaudrateComboItems()
        Form1.cmb_ComSpeed.Text = baudrate.ToString
        ' ---------------------------------------------------------------- prepare an array with the baudrate 
        Dim params(0) As Short
        params(0) = CShort(arrayindex)
        ' ---------------------------------------------------------------- send the message using all the valid baudrates 
        For i As Int32 = 0 To BaudRates.Length - 1
            '
            If i = arrayindex Then Continue For
            '
            COM_Open(PortName, BaudRates(i))
            If ComPort.IsOpen Then
                Threading.Thread.Sleep(10)
                SendFc16(0, regindex, 1, params)
            End If
        Next
    End Sub

    Friend Sub SetBaudrateComboItems()
        With Form1.cmb_ComSpeed
            Dim oldevents As Boolean = EventsAreEnabled
            EventsAreEnabled = False
            .Items.Clear()
            If BaudRates.Length > 0 Then
                For i As Int32 = 0 To BaudRates.Length - 1
                    .Items.Add(BaudRates(i).ToString)
                Next
            Else
                Dim b As Object() = {"1000000", "1000000", "1000000", "1000000", "1000000", "500000", _
                                     "256000", "250000", "128000", "115200", "74880", "57600", "56000", _
                                     "38400", "28800", "19200", "14400", "9600", "4800", "2400", "1200", "600"}
                .Items.AddRange(b)
            End If
            EventsAreEnabled = oldevents
        End With
    End Sub


    ' ===========================================================================
    '  BUILD MESSAGE
    ' ===========================================================================
    Private Sub BuildMessage(ByVal address As Byte, ByVal type As Byte, ByVal start As UInt16, ByVal registers As UInt16, ByRef message As Byte())
        '
        message(0) = address
        message(1) = type
        message(2) = CByte((start >> 8) And 255)
        message(3) = CByte(start And 255)
        message(4) = CByte((registers >> 8) And 255)
        message(5) = CByte(registers And 255)
        '
        ' ---------------------------------------- Array to receive CRC bytes
        Dim CRC As Byte() = New Byte(1) {}
        GetCRC(message, CRC)
        '
        message(message.Length - 2) = CRC(0)
        message(message.Length - 1) = CRC(1)
    End Sub


    ' ===========================================================================
    '  CHECK RESPONSE (CRC check)
    ' ===========================================================================
    Private Function CheckResponse(ByVal response As Byte()) As Boolean
        Dim CRC As Byte() = New Byte(1) {}
        GetCRC(response, CRC)
        If CRC(0) = response(response.Length - 2) AndAlso CRC(1) = response(response.Length - 1) Then
            Return True
        Else
            Return False
        End If
    End Function

    ' ===========================================================================
    '  GetCRC
    '   - Modbus message of any length 
    '   - 2 byte CRC array to return the CRC values
    ' ===========================================================================
    Private Sub GetCRC(ByVal message As Byte(), ByRef CRC As Byte())
        Dim CRCFull As UShort = &HFFFF
        Dim CRCLSB As Byte
        For i As Integer = 0 To message.Length - 3
            CRCFull = CUShort(CRCFull Xor message(i))
            For j As Integer = 0 To 7
                CRCLSB = CByte(CRCFull And &H1)
                CRCFull = CUShort((CRCFull >> 1) And &H7FFF)
                If CRCLSB = 1 Then
                    CRCFull = CUShort(CRCFull Xor &HA001)
                End If
            Next
        Next
        CRC(1) = CByte((CRCFull >> 8) And &HFF)
        CRC(0) = CByte(CRCFull And &HFF)
    End Sub

    ' =======================================================================================================
    '  GET RESPONSE
    ' =======================================================================================================
    ' Someone says that there is a bug in the DataReceived Event (that doesn't fire all of the time).  
    ' So we use the ReadByte command for a fixed length, which has proven reliable.
    ' =======================================================================================================
    '  CPU=6% FPS=240 (with 8 readings)
    'Private Sub GetResponse(ByRef response As Byte())
    '    For i As Integer = 0 To response.Length - 1
    '        Try
    '            response(i) = CByte(ComPort.ReadByte() And 255)
    '        Catch
    '        End Try
    '        If StopThreadFlag Then Return
    '    Next
    'End Sub

    '  CPU=0.5%  FPS=230  (with 8 readings)
    'Private Sub GetResponse(ByRef response As Byte())
    '    Do Until ComPort.BytesToRead >= response.Length
    '        If StopThreadFlag Then Return
    '        Threading.Thread.Sleep(1)
    '    Loop
    '    ComPort.Read(response, 0, response.Length)
    'End Sub

    ' --------------------------------------------------------------- THE BEST VERSION
    '  CPU=0.8% FPS=300 (with 8 readings)
    Private Sub GetResponse(ByRef response As Byte())
        COM_WaitDataReceived(response.Length)
        ComPort.Read(response, 0, response.Length)
    End Sub


    ' ===========================================================================
    '  Function 16 - Write Multiple Holding Registers
    ' ===========================================================================
    Private Function SendFc16(ByVal address As Byte, _
                              ByVal start As UInt16, _
                              ByVal registers As UInt16, _
                              ByVal values As Short()) As Boolean
        '
        If Not ComPort.IsOpen Then Return False
        ' -------------------------------------------------------------- Clear in/out buffers
        ComPort.DiscardOutBuffer()
        ComPort.DiscardInBuffer()
        ' -------------------------------------------------------------- Message is 1 addr + 1 fcn + 2 start + 2 reg + 1 count + 2 * reg vals + 2 CRC
        Dim message As Byte() = New Byte(9 + (2 * registers - 1)) {}
        ' -------------------------------------------------------------- Function 16 response is fixed at 8 bytes
        Dim response As Byte() = New Byte(7) {}
        ' -------------------------------------------------------------- Add bytecount to message
        message(6) = CByte(registers * 2)
        ' -------------------------------------------------------------- Put write values into message prior to sending
        For i As Integer = 0 To registers - 1
            message(7 + 2 * i) = CByte((values(i) >> 8) And 255)
            message(8 + 2 * i) = CByte(values(i) And 255)
        Next
        ' -------------------------------------------------------------- Build outgoing message
        BuildMessage(address, CByte(16), start, registers, message)
        ' -------------------------------------------------------------- Send Modbus message to Serial Port
        Try
            ComPort.Write(message, 0, message.Length)
            SentPackets += 1
            If address > 0 Then
                GetResponse(response)
                ' -------------------------------------------------------------- Evaluate message
                If CheckResponse(response) Then
                    ReceivedPackets += 1
                    Return True
                Else
                    Return False
                End If
            Else
                Threading.Thread.Sleep(5)
                Return True
            End If
        Catch err As Exception
            Return False
        End Try
    End Function

    ' ===========================================================================
    '  Function 3 - Read Multiple Holding Registers
    ' ===========================================================================
    Private Function SendFc3(ByVal address As Byte, _
                             ByVal start As UInt16, _
                             ByVal registers As UInt16, _
                             ByRef values As Short()) As Boolean
        '
        If ComPort.IsOpen Then
            ' -------------------------------------------------------------- Clear in/out buffers
            ComPort.DiscardOutBuffer()
            ComPort.DiscardInBuffer()
            ' -------------------------------------------------------------- Function 3 request is always 8 bytes
            Dim message As Byte() = New Byte(7) {}
            ' -------------------------------------------------------------- Function 3 response buffer
            Dim response As Byte() = New Byte(5 + (2 * registers - 1)) {}
            ' -------------------------------------------------------------- Build outgoing modbus message
            BuildMessage(address, CByte(3), start, registers, message)
            ' -------------------------------------------------------------- Send modbus message to Serial Port
            Try
                ComPort.Write(message, 0, message.Length)
                SentPackets += 1
                GetResponse(response)
            Catch err As Exception
                Return False
            End Try
            ' -------------------------------------------------------------- Evaluate message
            If CheckResponse(response) Then
                ' ---------------------------------------------------------- Return requested register values
                For i As Integer = 0 To (response.Length - 5) \ 2 - 1
                    values(i) = response(2 * i + 3)
                    values(i) <<= 8
                    values(i) += response(2 * i + 4)
                Next
                ReceivedPackets += 1
                Return True
            Else
                Return False
            End If
        Else
            Return False
        End If
    End Function

    ' ------------------------------------------------------------------------------------------
    ' Name	   Length (bits)	Function
    ' ------------------------------------------------------------------------------------------
    ' Start	    28	            At least 3½ character times of silence (mark condition)
    ' Address	8	            Station address
    ' Function	8	            Indicates the function code; e.g., read coils/holding registers
    ' Data	    n × 8	        Data + length will be filled depending on the message type
    ' CRC	    16	            Cyclic redundancy check
    ' End	    28	            At least 3½ character times of silence between frames


    ' -------------------------------------------------------------------------------------------
    ' Function name	                    Function code	Comment
    ' -------------------------------------------------------------------------------------------
    ' Read Discrete Inputs	             2	
    ' Read Coils	                     1	
    ' Write Single Coil	                 5	
    ' Write Multiple Coils	            15	
    ' Read Input Registers               4	
    ' Read Multiple Holding Registers	 3  <--         Implemented	
    ' Write Single Holding Register	     6	
    ' Write Multiple Holding Registers	16  <--         Implemented
    ' Read/Write Multiple Registers	    23	
    ' Mask Write Register	            22	
    ' Read FIFO Queue	                24	
    ' Read File Record	                20	
    ' Write File Record	                21	
    ' Read Exception Status	             7	            serial only
    ' Diagnostic	                     8	            serial only
    ' Get Com Event Counter	            11	            serial only
    ' Get Com Event Log                 12	            serial only
    ' Report Slave ID                   17	            serial only
    ' Read Device Identification        43	
    ' Encapsulated Interface Transport	43

End Module