﻿
Module Module_SaveLoad

    Friend EventsAreEnabled As Boolean = False
    'Friend Form1_VisibleAtStart As Boolean

    Friend GCI As Globalization.CultureInfo = Globalization.CultureInfo.InvariantCulture

    ' =======================================================================================================
    '   APP TITLE AND VERSION
    ' =======================================================================================================
    Friend Function AppTitleAndVersion(Optional ByVal Title As String = "") As String
        If Title = "" Then Title = Replace(My.Application.Info.AssemblyName, "_", " ")
        Dim s() As String = Split(My.Application.Info.Version.ToString, ".")
        AppTitleAndVersion = Title + " - V" + s(0) + "." + s(1)
        If s(2) <> "0" Then AppTitleAndVersion += "." + s(2)
        If s(3) <> "0" Then AppTitleAndVersion += "." + s(3)
    End Function


    ' =======================================================================================================
    '   Mixed functions
    ' =======================================================================================================

    ' ==============================================================
    '  REPLACE MULTIPLE SPACES and TABS and TRIM
    '  (3..5 uS for Strings with 32 chars)
    ' ============================================================== 
    Friend Function ReplaceMultipleSpacesAndTabAndTrim(ByVal s As String) As String
        s = s.Replace(vbTab, " ")
        While s.Contains("  ")
            s = s.Replace("  ", " ")
        End While
        Return s.Trim
    End Function

    ' =======================================================================================================
    '   Files
    ' =======================================================================================================
    ' returns lower-case extension with initial dot
    Public Function GetExtension(ByVal str As String) As String
        On Error Resume Next
        Return LCase(IO.Path.GetExtension(str))
    End Function
    Public Function RemoveExtension(ByVal str As String) As String
        On Error Resume Next
        Return PlatformAdjustedFileName(IO.Path.GetDirectoryName(str) & "\" & IO.Path.GetFileNameWithoutExtension(str))
    End Function
    Public Function FileExists(ByVal fileName As String) As Boolean
        Return My.Computer.FileSystem.FileExists(fileName)
    End Function
    Public Function FolderExists(ByVal FolderName As String) As Boolean
        If FolderName.Length < 2 Then Return False
        If OperatingSystemIsWindows Then
            FolderName = LCase(FolderName)
            Select Case FolderName
                Case "a:\", "b:\", "c:\", "d:\", "e:\", "f:\", "g:\", "h:\", "i:\", "j:\", "k:\"
                    Return True
            End Select
        End If
        Return My.Computer.FileSystem.DirectoryExists(FolderName)
    End Function


    ' =======================================================================================
    '  FORM
    ' =======================================================================================
    Friend Sub LimitFormPosition(ByVal f As System.Windows.Forms.Form)
        If f.WindowState <> FormWindowState.Normal Then Return
        GetMaxScreenBounds()
        EnsureFormVisible(f)
    End Sub
    Friend Sub LimitFormPosition_CompletelyVisible(ByVal f As System.Windows.Forms.Form)
        If f.WindowState <> FormWindowState.Normal Then Return
        GetMaxScreenBounds()
        EnsureFormCompletelyVisible(f)
    End Sub
    Private SB As Rectangle = New Rectangle(Integer.MaxValue, Integer.MaxValue, Integer.MinValue, Integer.MinValue)
    Private Sub GetMaxScreenBounds()
        For Each s As Screen In System.Windows.Forms.Screen.AllScreens
            SB = Rectangle.Union(SB, s.WorkingArea)
        Next
    End Sub
    Private Sub EnsureFormCompletelyVisible(ByVal frm As Form)
        With frm
            .Width = Math.Min(.Width, SB.Width)         ' not more than dimensions of a maximized window
            .Height = Math.Min(.Height, SB.Height)      ' not more than dimensions of a maximized window
            .Width = Math.Max(.Width, 32)               ' at least 32x24
            .Height = Math.Max(.Height, 24)             ' at least 32x24
            .Left = Math.Min(.Left, SB.Right - .Width)  ' not beyond the right border
            .Top = Math.Min(.Top, SB.Bottom - .Height)  ' not beyond the bottom border
            .Left = Math.Max(.Left, SB.Left)            ' at least at the left border
            .Top = Math.Max(.Top, SB.Top)               ' at least at the top border
        End With
    End Sub
    Private Sub EnsureFormVisible(ByVal frm As Form)
        With frm
            .Width = Math.Min(.Width, SB.Width)             ' not more than the VIRTUALSCREEN dimensions
            .Height = Math.Min(.Height, SB.Height)          ' not more than the VIRTUALSCREEN dimensions 
            .Width = Math.Max(.Width, 32)                   ' at least 32x24
            .Height = Math.Max(.Height, 24)                 ' at least 32x24
            .Left = Math.Min(.Left, SB.Right - 50)          ' not beyond the right border - 50 pixels
            .Top = Math.Min(.Top, SB.Bottom - 100)          ' not beyond the bottom border - 50 pixels
            .Left = Math.Max(.Left, SB.Left + 100 - .Width) ' at least at the left border + 50 pixels
            .Top = Math.Max(.Top, SB.Top - 10)              ' at least at the top border
        End With
    End Sub
    ' (The value of the RestoreBounds property is valid only 
    '   when the WindowState property of the Form class is not equal to Normal)
    Friend Function GetFormRectangle(ByVal frm As Form) As Rectangle
        Dim r As Rectangle
        If frm.WindowState = FormWindowState.Normal Then
            r = frm.Bounds
        Else
            r = frm.RestoreBounds
        End If
        Return r
    End Function


    ' ================================================================================================
    '  Private Read-Write functions
    ' ================================================================================================
    Private Function TabString(ByVal Name As String, _
                               Optional ByVal Value As Double = Double.NaN, _
                               Optional ByVal fmt As String = "") As String

        Dim nTab As Int32 = Math.Max(0, 22 - Name.Length)
        If Double.IsNaN(Value) Then
            Return Name
        Else
            Return Name & "=" & Strings.StrDup(nTab, " ") & Value.ToString(fmt, GCI)
        End If
    End Function
    Private Function TabString(ByVal Name As String, _
                                  ByVal Value As Boolean) As String

        Dim nTab As Int32 = Math.Max(0, 22 - Name.Length)

        Return Name & "=" & Strings.StrDup(nTab, " ") & Value.ToString
    End Function
    Private Function TabString(ByVal Name As String, _
                                ByVal Value As String) As String

        Dim nTab As Int32 = Math.Max(0, 22 - Name.Length)

        Return Name & "=" & Strings.StrDup(nTab, " ") & Value
    End Function
    Private Function Val_Single(ByVal l As String) As Single
        Return CSng(Val(l.Replace(",", ".")))
    End Function
    Private Function Val_Double(ByVal l As String) As Double
        Return Val(l.Replace(",", "."))
    End Function
    Private Function Val_Int(ByVal l As String) As Int32
        Return StringToInt32(l)
    End Function
    Friend Function ExtractParamName(ByRef s As String) As String
        ' ------------------------- Returns the first field from begin to the first "=" symbol
        ' -------------------------  and removes it from the string
        Dim i As Integer
        i = InStr(s, "=")
        If i > 0 Then
            ExtractParamName = Trim(Strings.Left(s, i - 1))
            s = Trim(Mid(s, i + 1))
        Else
            ExtractParamName = Trim(s)
            s = ""
        End If
    End Function
    Friend Function AssemblyName() As String
        Return System.Reflection.Assembly.GetExecutingAssembly.GetName.Name
    End Function


    ' ==================================================================================================
    '  SAVE LOAD -- Program INI
    ' ==================================================================================================
    Friend Sub Save_INI()
        Dim iniFileName As String = PlatformAdjustedFileName(Application.StartupPath & "\" & AssemblyName() & "_INI.txt")
        Dim f As System.IO.StreamWriter = Nothing
        Try
            f = IO.File.CreateText(iniFileName)
            '
            f.WriteLine("")
            f.WriteLine("==============================================")
            f.WriteLine(" Properties")
            f.WriteLine("==============================================")
            ' ------------------------------------------------------------------------------ FORM BOUNDS
            Dim r As Rectangle
            r = GetFormRectangle(Form1)
            f.WriteLine(TabString("Form1_Top", r.Top))
            f.WriteLine(TabString("Form1_Left", r.Left))
            f.WriteLine(TabString("Form1_Width", r.Width))
            f.WriteLine(TabString("Form1_Height", r.Height))
            f.WriteLine(TabString("Form1_WindowState", Form1.WindowState))
            'f.WriteLine(TabString("Form1_VisibleAtStart", Form1.Visible And Form1.WindowState <> FormWindowState.Minimized))
            ' ------------------------------------------------------------------------------
            r = GetFormRectangle(Form_Input)
            f.WriteLine(TabString("Form_Input_Top", Form_Input.FormInputLocation.Y))
            f.WriteLine(TabString("Form_Input_Left", Form_Input.FormInputLocation.X))
            '
            f.WriteLine("")
            f.WriteLine("==============================================")
            f.WriteLine(" Options")
            f.WriteLine("==============================================")
            f.WriteLine(TabString("Zoom", Form1.tkbar_Zoom.Value))
            f.WriteLine(TabString("Transparency", Form1.tkbar_Transparency.Value))
            f.WriteLine(TabString("ID_Enable", Form_Options.NumericUpDown_Enable.Value))
            f.WriteLine(TabString("ID_Address", Form_Options.NumericUpDown_Address.Value))
            f.WriteLine(TabString("ID_Value", Form_Options.NumericUpDown_Identifier.Value))
            '
            f.WriteLine("")
            f.WriteLine("==============================================")
            f.WriteLine(" Communication options")
            f.WriteLine("==============================================")
            f.WriteLine(TabString("ComPort", Form1.cmb_ComPort.Text))
            f.WriteLine(TabString("ComSpeed", Form1.cmb_ComSpeed.Text))
            f.WriteLine(TabString("Reconnect", Form1.chk_Reconnect.Checked))
            f.WriteLine(TabString("ScanPorts", Form1.chk_ScanPorts.Checked))
            '
            f.WriteLine(TabString("CommandsFile", CommandsFileName))
        Catch
        End Try
        Try
            f.Close()
        Catch
        End Try
    End Sub

    Friend Sub Load_INI()
        ' ------------------------------------------------------------------------------- 
        ' With "Resume Next" subsequent parameters are loaded and f.Close() is executed
        ' -------------------------------------------------------------------------------
        On Error Resume Next  ' use Resume-Next instead of Try-Catch
        ' -------------------------------------------------------------------------------
        Dim l As String
        Dim iniFileName As String = PlatformAdjustedFileName(Application.StartupPath & "\" & AssemblyName() & "_INI.txt")
        If My.Computer.FileSystem.FileExists(iniFileName) Then
            '
            Dim f As System.IO.StreamReader
            f = IO.File.OpenText(iniFileName)
            '
            Do While Not f.EndOfStream
                l = f.ReadLine()
                Select Case ExtractParamName(l)
                    Case "Form1_Top" : Form1.Top = Val_Int(l)
                    Case "Form1_Left" : Form1.Left = Val_Int(l)
                    Case "Form1_Width" : Form1.Width = Val_Int(l)
                    Case "Form1_Height" : Form1.Height = Val_Int(l)
                    Case "Form1_WindowState" : Form1.WindowState = CType((Val(l)), FormWindowState) : Form1.Refresh()
                        'Case "Form1_VisibleAtStart" : Form1_VisibleAtStart = l = "True"
                        '
                        ' ---------------------------------------------------------------
                    Case "Form_Input_Top" : Form_Input.Top = Val_Int(l)
                    Case "Form_Input_Left" : Form_Input.Left = Val_Int(l)
                        '
                    Case "Zoom" : Form1.tkbar_Zoom.Value = Val_Int(l)
                    Case "Transparency" : Form1.tkbar_Transparency.Value = Val_Int(l)
                    Case "ID_Enable" : Form_Options.NumericUpDown_Enable.Value = Val_Int(l)
                    Case "ID_Address" : Form_Options.NumericUpDown_Address.Value = Val_Int(l)
                    Case "ID_Value" : Form_Options.NumericUpDown_Identifier.Value = Val_Int(l)
                        '
                    Case "ComPort" : Form1.cmb_ComPort.Text = l
                    Case "ComSpeed" : Form1.cmb_ComSpeed.Text = l
                    Case "Reconnect" : Form1.chk_Reconnect.Checked = l = "True"
                    Case "ScanPorts" : Form1.chk_ScanPorts.Checked = l = "True"
                        '
                    Case "CommandsFile" : CommandsFileName = l
                End Select
            Loop
            '
            ' ---------------------------------------------- 
            f.Close()
        End If
        '
        LimitFormPosition(Form1)
    End Sub

    ' =========================================================================================================
    '  RTB-TEXT SAVE/LOAD
    ' =========================================================================================================
    Friend FilesPath As String = Application.StartupPath + "\Files"
    Friend CommandsFileName As String

    Friend Sub SaveCommandsFile()
        Dim fname As String = FilesPath + "\" + CommandsFileName
        If Not IO.File.Exists(fname) Then
            IO.Directory.CreateDirectory(FilesPath)
            CommandsFileName = IO.Path.GetFileName(CommandsFileName)
            If CommandsFileName = "" Then CommandsFileName = "TestFile1.txt"
            fname = FilesPath + "\" + CommandsFileName
            Form1.GroupBox1.Text = IO.Path.GetFileNameWithoutExtension(CommandsFileName)
        End If
        RTB_Lines = Form1.RTB.Lines
        IO.File.WriteAllLines(fname, RTB_Lines, System.Text.Encoding.Unicode)
    End Sub

    Friend Sub LoadFileDialog()
        Dim ofd As OpenFileDialog = New OpenFileDialog()
        ofd.InitialDirectory = FilesPath
        ofd.FileName = IO.Path.GetFileName(CommandsFileName)
        ofd.AutoUpgradeEnabled = True
        ofd.Title = "Load file"
        ofd.ValidateNames = False
        ofd.Multiselect = True
        ofd.Filter = "All files (*.*)|*.*"
        ofd.DefaultExt = ".txt"
        '
        SendKeys.Send("+{TAB}+{TAB}" + ofd.FileName + "{TAB}{TAB}") ' file selected in the list
        ''SendKeys.Send("{HOME}{END}^A") ' name visible and selected
        SendKeys.Send("{HOME}{END}")   ' name visible and unselected
        '
        IO.Directory.CreateDirectory(ofd.InitialDirectory)
        If ofd.ShowDialog() = Windows.Forms.DialogResult.OK Then
            Dim fpath As String = IO.Path.GetDirectoryName(ofd.FileName)
            If fpath = FilesPath Then
                CommandsFileName = IO.Path.GetFileName(ofd.FileName)
            Else
                MessageBox.Show("The selected file is not in the ""Files"" folder." + vbCrLf + vbCrLf + _
                                "Please select a file in the ""Files"" folder.")
            End If
        End If
    End Sub

    Private Function GetPath(ByVal str As String) As String
        Try
            Return IO.Path.GetDirectoryName(str) & "\"
        Catch
            Return str
        End Try
    End Function

    Private Const InvalidUnicodes As String = "[\u0590-\u097F]|" + _
                                              "\u200E|\u200F|\u202A|\u202B|\u202C|\u202D|\u202E"

    Friend Sub LoadFile()
        Dim fname As String = FilesPath + "\" + CommandsFileName
        '
        If FileExists(fname) Then
            ' ----------------------------------------------- load as unicode (or any other format specified in the txt)
            Dim sa As String() = IO.File.ReadAllLines(fname)
            ' ----------------------------------------------- trim end chars and remove unwanted unicode chars
            For i As Int32 = 0 To sa.Length - 1
                sa(i) = sa(i).TrimEnd
                ' -------------------------------------------------------------------------------
                '  Remove the unwanted unicode chars 
                ' -------------------------------------------------------------------------------
                sa(i) = System.Text.RegularExpressions.Regex.Replace(sa(i), InvalidUnicodes, "")
                ' -------------------------------------------------------------------------------
            Next
            '
            Form1.RTB.Enabled = False
            Form1.RTB.PaintEventEnabled = False
            '
            Form1.RTB.Lines = sa
            RTB_Lines = Form1.RTB.Lines
            Form1.RTB.InitProps()
            Form1.RTB.ColorAllLines()
            Form1.RTB.Undo_Clear()
            Form1.RTB.History_Clear()
            Form1.RTB.ShowLinesCount()
            Form1.ShowHideTextAndTools()

            Form1.GroupBox1.Text = IO.Path.GetFileNameWithoutExtension(CommandsFileName)
            'Form1.GroupBox1.BackColor = Color.LemonChiffon
            '
            Form1.RTB.PaintEventEnabled = True
            Form1.RTB.Enabled = True
        End If
    End Sub

End Module
