﻿
Imports System.Net
Imports System.Collections.Generic

Module TransferFunctions

    Friend LoopTimeMillisec As Int32 = 1
    Friend SectionSelectorSlot As Int32 = -1
    Friend ActualSection As Int32
    Friend ActualSectionSlotValue As Int32

    Private TransferData_List As New List(Of TransferData)

    Friend Enum TransferTypes
        SlotToModbus
        ModbusToSlot
        ImmediateToModbus
    End Enum

    Friend Class TransferData
        Public TransferType As TransferTypes
        Public Slot As Int32
        Public NumSlots As UInt16
        Public Device As Byte
        Public Address As UInt16
        Public RegisterType As Byte
        Public OldValues() As Int16
        Public ImmediateValue As Int32
        Public Executed As Boolean = False
        Public Sub New(ByVal TransferType As TransferTypes, ByVal Slot As Int32, _
                                                            ByVal NumSlots As Int32, _
                                                            ByVal Device As Int32, _
                                                            ByVal Address As UInt16, _
                                                            ByVal RegisterType As Byte, _
                                                            Optional ByVal ImmediateValue As Int32 = 0)
            Me.TransferType = TransferType
            Me.Slot = Slot
            Me.NumSlots = CUShort(NumSlots)
            Me.Device = CByte(Device)
            Me.Address = Address
            Me.RegisterType = RegisterType
            ReDim Me.OldValues(NumSlots - 1)
            For i As Int32 = 0 To NumSlots - 1
                Me.OldValues(i) = Short.MinValue
            Next
            Me.ImmediateValue = ImmediateValue
        End Sub
    End Class

    ' ===========================================================================================================
    '   REINIT TRANSFER LISTS
    ' ===========================================================================================================
    Friend Sub ReinitTransferLists(ByVal la() As String)
        Dim sa As String()
        ' ------------------------------------------------------------------- Read SectionSelectorSlot and looptimemillisec before 
        SectionSelectorSlot = -1 ' disabled by default
        ReDim BaudRates(-1)
        BaudRateRegisterIndex = -1
        ActualSection = 0
        '
        For Each l As String In la
            ' --------------------------------------------------------------- Tokenize line
            sa = Form1.RTB.Tokenize(l)
            ' --------------------------------------------------------------- Decode line
            Select Case sa.Length
                ' ----------------------------------------------------------- Commands
                Case 2
                    Select Case sa(0)
                        Case "baudrates"
                            If ValidLine_BaudRates(sa) Then
                                BaudRates = GetBaudRatesArray(sa)
                            End If
                        Case "baudrateregister"
                            If ValidLine_BaudRateRegister(sa) Then
                                BaudRateRegisterIndex = StringToInt32(sa(1))
                            End If
                        Case "sectionselectorslot"
                            If ValidSectionSelectionSlot(sa(1)) Then
                                SectionSelectorSlot = StringToInt32(sa(1))
                            End If
                        Case "looptimemillisec"
                            If ValidLoopTimeMillisec(sa(1)) Then
                                LoopTimeMillisec = StringToInt32(sa(1))
                            End If
                    End Select
            End Select
        Next
        '
        ' ------------------------------------------------------------------- Read MBUS commands
        TransferData_List.Clear()
        For Each l As String In la
            l = ReplaceMultipleSpacesAndTabAndTrim(l).ToLower
            Form1.RTB.RemoveComments(l)
            sa = l.Split(" "c)
            ' --------------------------------------------------------------- Test section
            If SectionSelectorSlot > 0 Then
                If sa.Length = 2 AndAlso sa(0) = "section" Then
                    ActualSection = StringToInt32(sa(1))
                End If
                If ActualSection <> ActualSectionSlotValue And ActualSection <> 0 Then
                    Continue For
                End If
            End If
            ' --------------------------------------------------------------- Decode line
            Select Case sa.Length
                ' -----------------------------------------------------------
                Case 1
                    Select Case sa(0)
                        Case "end"
                            Exit For
                    End Select
                Case 5
                    ' -------------------------------------------------------
                    Select Case sa(0)
                        Case "   "
                            ' free for eventual other commands with 5 elements
                        Case Else
                            If ValidLine_ImmediateToModBus(sa) Then
                                TransferData_List.Add(New TransferData(TransferTypes.ImmediateToModbus, _
                                                                       -1, _
                                                                       0, _
                                                                       DecodeDevice(sa(3)), _
                                                                       DecodeAddress(sa(4)), _
                                                                       RegisterType, _
                                                                       StringToInt32(sa(0))))
                            End If
                    End Select
                Case 6
                    ' ------------------------------------------------------- SlotToModBus / ModBusToSlot
                    Select Case sa(0)
                        Case "slot"
                            If ValidLine_SlotToModBus(sa) Then
                                TransferData_List.Add(New TransferData(TransferTypes.SlotToModbus, _
                                                                        DecodeSlot(sa(1)), _
                                                                        1, _
                                                                        DecodeDevice(sa(4)), _
                                                                        DecodeAddress(sa(5)), _
                                                                        RegisterType))
                            End If
                        Case "mbus", "modbus"
                            If ValidLine_ModBusToSlot(sa) Then
                                TransferData_List.Add(New TransferData(TransferTypes.ModbusToSlot, _
                                                                        DecodeSlot(sa(5)), _
                                                                        1, _
                                                                        DecodeDevice(sa(1)), _
                                                                        DecodeAddress(sa(2)), _
                                                                        RegisterType))
                            End If
                    End Select
                Case 7
                    ' ------------------------------------------------------- SlotToModBus multiple / ModBusToSlot multiple
                    Select Case sa(0)
                        Case "slot"
                            If ValidLine_SlotToModBus_Multiple(sa) Then
                                TransferData_List.Add(New TransferData(TransferTypes.SlotToModbus, _
                                                                        DecodeSlot(sa(1)), _
                                                                        DecodeSlot(sa(2)) - DecodeSlot(sa(1)) + 1, _
                                                                        DecodeDevice(sa(5)), _
                                                                        DecodeAddress(sa(6)), _
                                                                        RegisterType))
                            End If
                        Case "mbus", "modbus"
                            If ValidLine_ModBusToSlot_Multiple(sa) Then
                                TransferData_List.Add(New TransferData(TransferTypes.ModbusToSlot, _
                                                                        DecodeSlot(sa(5)), _
                                                                        DecodeSlot(sa(6)) - DecodeSlot(sa(5)) + 1, _
                                                                        DecodeDevice(sa(1)), _
                                                                        DecodeAddress(sa(2)), _
                                                                        RegisterType))
                            End If
                    End Select
            End Select
        Next
    End Sub

    Friend Sub UpdateSectionSelector()
        If SectionSelectorSlot > 0 Then
            Dim n As Int32 = CInt(Slots.ReadSlot(SectionSelectorSlot))
            If n <> ActualSectionSlotValue Then
                ActualSectionSlotValue = n
                RTB_Lines = Form1.RTB.Lines
                ReinitTransferLists(RTB_Lines)
            End If
        End If
    End Sub

    Friend Function DecodeSlot(ByVal s As String) As Int32
        Return StringToInt32(s)
    End Function

    Friend Function DecodeDevice(ByVal s As String) As Int32
        Return StringToInt32(s)
    End Function

    Private RegisterType As Byte
    Friend Function DecodeAddress(ByVal s As String) As UInt16
        Select Case s(0)
            Case "0"c : RegisterType = 0 ' Read Write -  1 bit - Digital output (Discrete input  / Coil)
            Case "1"c : RegisterType = 1 ' Read Only  -  1 bit - Digital input  (Discrete output / Contact)
            Case "3"c : RegisterType = 3 ' Read Only  - 16 bit - Analog input   (Input register / ADC)
            Case "4"c : RegisterType = 4 ' Read Write - 16 bit - Register       (Read-Write Register)
            Case Else
                RegisterType = 255
                Return 0
        End Select
        s = s.Substring(1)
        Return CUShort(Val(s) - 1)
    End Function

    ' ===========================================================================================================
    '   READ WRITE REGISTERS
    ' ===========================================================================================================
    Friend Sub ReadWriteRegisters()
        If Not ComPort.IsOpen Then Return
        Try
            For Each tdata As TransferData In TransferData_List
                Select Case tdata.TransferType
                    Case TransferTypes.SlotToModbus
                        ' --------------------------------------------------------------------------------------- TRANSFER SLOT to MODBUS
                        Select Case tdata.RegisterType
                            Case 0
                            Case 1
                            Case 3
                            Case 4
                                Dim Changed As Boolean = False
                                Dim values(tdata.NumSlots - 1) As Short
                                For i As Int32 = 0 To tdata.NumSlots - 1
                                    values(i) = CShort(Slots.ReadSlot_NoNan(tdata.Slot + i))
                                    If values(i) <> tdata.OldValues(i) Then
                                        tdata.OldValues(i) = values(i)
                                        Changed = True
                                    End If
                                    If StopThreadFlag Then Return
                                Next
                                If Changed Then
                                    ModBus_WriteRegister(tdata.Device, tdata.Address, tdata.NumSlots, values)
                                End If
                        End Select
                    Case TransferTypes.ModbusToSlot
                        ' --------------------------------------------------------------------------------------- TRANSFER MODBUS to SLOT 
                        Select Case tdata.RegisterType
                            Case 0
                            Case 1
                            Case 3
                            Case 4
                                Dim values(tdata.NumSlots - 1) As Short
                                values = ModBus_GetRegister(tdata.Device, tdata.Address, tdata.NumSlots)
                                For i As Int32 = 0 To tdata.NumSlots - 1
                                    Slots.WriteSlot(tdata.Slot + i, values(i))
                                    If StopThreadFlag Then Return
                                Next
                        End Select
                        ' --------------------------------------------------------------------------------------- TRANSFER IMMEDIATE TO MODBUS
                    Case TransferTypes.ImmediateToModbus
                        If Not tdata.Executed Then
                            Modbus_ImmediateToModbus(tdata)
                            tdata.Executed = True
                        End If
                        '
                End Select
            Next
            ' --------------------------------------------------------------------------------------- reduce CPU charge
            If ReceivedPackets = 0 Then
                Threading.Thread.Sleep(1)
            End If
            ' --------------------------------------------------------------------------------------- Update counters
            ExchangedFrames += 1
        Catch
            'Beep()
        End Try
    End Sub


    ' ===========================================================================================================
    '   COMMUNICATION STATISTICS
    ' ===========================================================================================================
    Friend Stat_Working As Double
    Friend Stat_FPS As Double
    Private ExchangedFrames As Int32

    Friend Sub Statistics_Update()
        ' ---------------------------------------------------------------- Working or Inactive
        If SentPackets = 0 Then
            Stat_Working -= 1
            If Stat_Working < 0 Then Stat_Working = 0
        Else
            Stat_Working += 1
            If Stat_Working > 5 Then Stat_Working = 5
        End If
        ' ---------------------------------------------------------------- measure the true milliseconds
        Static SW As Stopwatch = New Stopwatch
        Dim millisec As Double = SW.Elapsed.TotalMilliseconds
        If SW.IsRunning And millisec = 0 Then Return
        SW.Reset()
        SW.Start()
        ' ---------------------------------------------------------------- update Stat_FPS 
        If millisec > 0 Then
            SmoothValue_Pow_Adaptive(Stat_FPS, ExchangedFrames * 1000 / millisec, 10)
        End If
        ' ---------------------------------------------------------------- reset all values
        SentPackets = 0
        ReceivedPackets = 0
        ExchangedFrames = 0
        '
    End Sub


    ' ===========================================================================================================
    '   FAST TIMER
    ' ===========================================================================================================
    Private FastTimerThread As Threading.Thread
    Friend StopThreadFlag As Boolean

    Friend Sub FastTimer_Start()
        FastTimer_Stop()
        ' --------------------------------------------------------- 
        SchedulerMaxPrecision()
        ' --------------------------------------------------------- start the Thread 
        FastTimerThread = New Threading.Thread(AddressOf FastThread_Execute)
        ' --------------------------------------------------------- make the thread as background thread.
        FastTimerThread.IsBackground = True
        ' --------------------------------------------------------- set the Priority of the thread.
        FastTimerThread.Priority = Threading.ThreadPriority.AboveNormal
        ' --------------------------------------------------------- start the thread
        StopThreadFlag = False
        FastTimerThread.Start()
    End Sub

    Friend Sub FastTimer_Stop()
        If FastTimerThread Is Nothing Then Return
        If Not FastTimerThread.IsAlive Then Return
        ' --------------------------------------------- try to join the thread
        COM_WaitDataReceived_Stop()
        StopThreadFlag = True
        FastTimerThread.Join(10)
        ' --------------------------------------------- catch because sometimes the thread is already closed
        Try
            While FastTimerThread.IsAlive
                Threading.Thread.Sleep(10)
                Application.DoEvents()
            End While
            ' ----------------------------------------- eventually abort
            If FastTimerThread.IsAlive Then
                MessageBox.Show("ExecutionThread not closed correctly.", _
                                "Message from Theremino-ModBus", _
                                MessageBoxButtons.OK, MessageBoxIcon.Warning)
                FastTimerThread.Abort()
            End If
        Catch
            ' ----------------------------------------- try catch needed to be sure to close the app
            Try
                FastTimerThread.Abort()
            Catch
            End Try
        End Try
        ' ---------------------------------------------
        FastTimerThread = Nothing
        ' ---------------------------------------------
        SchedulerDefaultPrecision()
    End Sub

    Private Sub FastThread_Execute()
        Do
            ReadWriteRegisters()
            Threading.Thread.Sleep(LoopTimeMillisec)
        Loop Until StopThreadFlag
    End Sub

    ' =========================================================================
    '  WINDOWS SCHEDULER PRECISION
    ' ========================================================================= 
    Private Declare Function timeBeginPeriod Lib "winmm.dll" (ByVal uPeriod As Int32) As Int32
    Private Declare Function timeEndPeriod Lib "winmm.dll" (ByVal uPeriod As Int32) As Int32
    Private Sub SchedulerMaxPrecision()
        If OperatingSystemIsWindows Then
            timeBeginPeriod(1)
        End If
    End Sub
    Private Sub SchedulerDefaultPrecision()
        If OperatingSystemIsWindows Then
            timeEndPeriod(1)
        End If
    End Sub

End Module